@@ -5,10 +5,12 @@ import base64
 import binascii
 import blake3
 import datetime
-import hashlib
 import dockerfile
+import fasteners
+import hashlib
 import io
 import json
+import multiprocessing
 import os
 import paramiko
 import requests
@@ -29,7 +31,6 @@ VISOR_DIR = os.path.expanduser('~/.visor')
 VISOR_CONFIG = os.path.join(VISOR_DIR, 'config.json')
 
 # TODO: make ctrl+c properly cancel a build
-# TODO: don't take snapshots of very fast commands
 # TODO: partition by team
 # TODO: should be able to `visor console <container_id>` to recover from broken ssh
 # TODO: REPEATABLE should be taking state of all required machines not just cur
@@ -39,6 +40,27 @@ VISOR_CONFIG = os.path.join(VISOR_DIR, 'config.json')
 # TODO: host CPU going crazy even when guest cpu is close to idle
 # TODO: server should be able to specify root dir instead of ~/.visorserver
 
+def run_as_process(func, *args):
+  queue = multiprocessing.Queue()
+
+  def wrap_for_subprocess(func, *args):
+    queue.put(func(*args))
+
+  p = multiprocessing.Process(target=wrap_for_subprocess, args=[func, *args])
+
+  try:
+    p.start()
+    while queue.empty():
+      if not p.is_alive():
+        exit(1)
+      time.sleep(0.1)
+    ret = queue.get()
+    p.join()
+  finally:
+    p.terminate()
+  queue.close()
+  return ret
+
 config = {}
 if os.path.exists(VISOR_CONFIG):
   with open(VISOR_CONFIG, 'r') as f:
@@ -193,6 +215,17 @@ def find_context_files(visorfile):
               add_all_files_and_subfiles(
                 context_dir, files, os.path.join(visorfile_dir, host_path)
               )
+        elif command.cmd == 'repeatable':
+          values = get_command_value(command).split(' ')
+          if len(values) == 3 and values[0].lower() == 'script':
+            if values[2].startswith('/'):
+              add_all_files_and_subfiles(
+                context_dir, files, os.path.join(context_dir, values[2])
+              )
+            else:
+              add_all_files_and_subfiles(
+                context_dir, files, os.path.join(visorfile_dir, values[2])
+              )
     if not next_visorfile:
       break
     cur_visorfile = next_visorfile
@@ -267,7 +300,7 @@ def chunkify(context_dir, files):
       chunks[chunk.hash] = chunk.data
   return filemap, chunks
 
-def transmit_context(visorfile, archive):
+def subprocess_transmit_context(visorfile):
   files = find_context_files(visorfile)
   context_dir = find_context_dir(visorfile)
   filemap, chunks = chunkify(context_dir, files)
@@ -288,6 +321,22 @@ def transmit_context(visorfile, archive):
     )
   return filemap
 
+def transmit_context(visorfile):
+  # HACK: This part of the code needs to keep "fat" cdc chunks in memory.
+  # In the worst case, this is the size of the entire repository being uploaded.
+  # This can cause OOM if we have many clients run at nearly the same time,
+  # such as starting multiple runs from a ci process.
+  # In order to prevent this, we have a global lock on uploads that ends with
+  # a forced garbage collect of the chunks. However, python's garbage collector
+  # does not work as expected in this case. So to get around this, we put the
+  # chunk processing in a subprocess (the chunks are freed when the subprocess
+  # exits.
+  # In most cases this lock is held for less than a second.
+  # If uploading a large repository for the first time,
+  # it will be held for the duration of the upload.
+  with fasteners.InterProcessLock('/var/lock/visor-client-upload.lock'):
+    return run_as_process(subprocess_transmit_context, visorfile)
+
 def do_build_base(args):
   safe_stream_api_call(
     'post',
@@ -297,19 +346,13 @@ def do_build_base(args):
     },
   )
 
-def do_build(args):
-  if not args.snapshot_on_failure:
-    args.snapshot_on_failure = []
-  for subargs in args.snapshot_on_failure:
-    subargs[1] = int(subargs[1])
-  nonce = binascii.b2a_hex(os.urandom(15)).decode('utf8')
-  archive = '%s-%s' % (datetime.datetime.now().strftime('%Y%m%d'), nonce)
+def send_streaming_command(args, command, json_data):
   print('Sending build context...')
   server_url = verify_login()
   visorfile = args.visorfile
   if os.path.isdir(visorfile):
     visorfile = os.path.join(visorfile, 'Visorfile')
-  filemap = transmit_context(visorfile, archive)
+  filemap = transmit_context(visorfile)
   print('Done sending build context\n')
 
   buildenvs = {}
@@ -318,25 +361,53 @@ def do_build(args):
     buildenvs[a] = b
 
   context_dir = find_context_dir(visorfile)
-  safe_stream_api_call(
-    'post',
-    'build',
-    json={
-      'archive': archive,
+  json_data.update({
       'visorfileRelpath': os.path.relpath(visorfile, start=context_dir),
-      'keepRunning': args.keep_running,
-      'snapshotOnFailure': args.snapshot_on_failure,
       'buildenvs': buildenvs,
       # TODO: transmitting the filemap once here and once above in the
       # transmitmeta. Could keep something on the server so we only have to
       # transmit once. The file metadata can get pretty big for lots of files.
       'filemap': filemap,
-    },
+  })
+
+  safe_stream_api_call(
+    'post',
+    command,
+    json=json_data,
   )
 
+
+def do_build(args):
+  if not args.snapshot_on_failure:
+    args.snapshot_on_failure = []
+  for subargs in args.snapshot_on_failure:
+    subargs[1] = int(subargs[1])
+  send_streaming_command(args, 'build', {
+    'keepRunning': args.keep_running,
+    'snapshotOnFailure': args.snapshot_on_failure,
+  })
+
+
+def do_multirun(args):
+  send_streaming_command(args, 'multirun', {
+    'numRuns': args.num,
+    'passString': args.pass_string,
+    'failString': args.fail_string,
+  })
+
+
+def do_log(args):
+  resp = _api_call('get', f'result/{args.multirun_id}?run={args.run_num}')
+  if resp.status_code == 200:
+    print(resp.text)
+    return
+    # return resp.json()
+  raise Exception('Server responded with %s: %s' % (resp.status_code, resp.text))
+  # print(safe_non_stream_api_call())
+
 def do_run(args):
   resp = safe_non_stream_api_call(
-    'post', 'run', json={'imageId': args.image_id}
+    'post', 'run', json={'imageId': args.image_id}, timeout=600
   )
   print(resp['containerId'])
 
@@ -366,7 +437,7 @@ def do_ssh(args, server_url=None):
     ] if x]
     os.execvp('ssh', cmd)
 
-def do_console(args):
+def do_console(args, server_url=None):
   guest_console_args = safe_non_stream_api_call(
     'get', 'guest_console_args', json={'containerId': args.container_id}
   )
@@ -378,6 +449,16 @@ def do_console(args):
     subcmd = 'telnet localhost {port}'
     # TODO: LOH not sure how to make telnet secure
 
+def do_throttle(args):
+  json_to_send = {
+    'containerId': args.container_id
+  }
+  if args.disk_bps is not None:
+    json_to_send['diskBps'] = args.disk_bps
+  if args.disk_iops is not None:
+    json_to_send['diskIops'] = args.disk_iops
+  resp = safe_non_stream_api_call('post', 'throttle', json=json_to_send)
+
 def do_ps(args):
   resp = safe_non_stream_api_call('get', 'ps')
   if args.quiet:
@@ -491,6 +572,19 @@ def main():
   build_parser.add_argument('visorfile', help='The Visorfile to build')
   build_parser.set_defaults(func=do_build)
 
+  multirun_parser = subparsers.add_parser('multirun', help='Build a Visorfile multiple time')
+  multirun_parser.add_argument('--build-env', dest='buildenvs', action='append', default=[])
+  multirun_parser.add_argument('visorfile', help='The Visorfile to build')
+  multirun_parser.add_argument('num', type=int, default=10, help='The number of time the visorfile needs to be run')
+  multirun_parser.add_argument('--pass-string', type=str, default='', help='The string to search for pass status')
+  multirun_parser.add_argument('--fail-string', type=str, default='', help='The string to search for fail status')
+  multirun_parser.set_defaults(func=do_multirun)
+
+  log_parser = subparsers.add_parser('log', help='Download a log from multiplerun command')
+  log_parser.add_argument('multirun_id', type=str, help='The id of the multirun')
+  log_parser.add_argument('run_num', type=str, nargs="?", default="", help='The run number of the log')
+  log_parser.set_defaults(func=do_log)
+
   run_parser = subparsers.add_parser('run', help='Run an instance of image_id')
   run_parser.add_argument('image_id', type=str)
   run_parser.set_defaults(func=do_run)
@@ -506,6 +600,14 @@ def main():
   console_parser.add_argument('container_id', type=str)
   console_parser.set_defaults(func=do_console)
 
+  throttle_parser = subparsers.add_parser(
+    'throttle', help='Throttle the disk of a running container'
+  )
+  throttle_parser.add_argument('container_id', type=str)
+  throttle_parser.add_argument('--disk-bps', type=int)
+  throttle_parser.add_argument('--disk-iops', type=int)
+  throttle_parser.set_defaults(func=do_throttle)
+
   ps_parser = subparsers.add_parser('ps', help='List running containers')
   ps_parser.add_argument('--quiet', '-q', action='store_true')
   ps_parser.set_defaults(func=do_ps)
